==> hlog-0.14.tar.gz: hlog-0.14/qso_serde.c <==
/*
 * Copyright (c) 2020 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <jeffpc/error.h>
#include <jeffpc/nvl.h>

#include "qso_impl.h"
#include "version.h"

#define QSO_VERSION	1

static int nvl_set_time(struct nvlist *nvl, const char *name,
			const struct time *time)
{
	struct val *arr[6];

	if (time->have & TIME_PART_DATE) {
		arr[0] = VAL_ALLOC_INT(time->year);
		arr[1] = VAL_ALLOC_INT(time->month);
		arr[2] = VAL_ALLOC_INT(time->day);
	} else {
		arr[0] = VAL_ALLOC_NULL();
		arr[1] = VAL_ALLOC_NULL();
		arr[2] = VAL_ALLOC_NULL();
	}

	if (time->have & TIME_PART_TIME_HM) {
		arr[3] = VAL_ALLOC_INT(time->hour);
		arr[4] = VAL_ALLOC_INT(time->min);
	} else {
		arr[3] = VAL_ALLOC_NULL();
		arr[4] = VAL_ALLOC_NULL();
	}

	if (time->have & TIME_PART_TIME_S)
		arr[5] = VAL_ALLOC_INT(time->sec);
	else
		arr[5] = VAL_ALLOC_NULL();

	return nvl_set_array_copy(nvl, name, arr, ARRAY_LEN(arr));
}

static struct nvlist *get_part(const struct qso_side *side)
{
	struct nvlist *nvl;
	int ret;

	nvl = nvl_alloc();
	if (IS_ERR(nvl))
		return nvl;

	ret = nvl_set_str(nvl, "station", str_getref(side->station_call));
	if (ret)
		goto err;

	ret = nvl_set_str(nvl, "operator", str_getref(side->operator_call));
	if (ret)
		goto err;

	ret = nvl_set(nvl, "power", val_getref(side->power));
	if (ret)
		goto err;

	ret = nvl_set(nvl, "freq", val_getref(side->freq));
	if (ret)
		goto err;

	ret = nvl_set_str(nvl, "mode", str_getref(side->mode));
	if (ret)
		goto err;

	ret = nvl_set_str(nvl, "rst", str_getref(side->rst));
	if (ret)
		goto err;

	ret = nvl_set_str(nvl, "name", str_getref(side->name));
	if (ret)
		goto err;

	ret = nvl_set_str(nvl, "qth", str_getref(side->qth));
	if (ret)
		goto err;

	ret = nvl_set_str(nvl, "grid", str_getref(side->grid));
	if (ret)
		goto err;

	ret = nvl_set_str(nvl, "country", str_getref(side->country));
	if (ret)
		goto err;

	ret = nvl_set_str(nvl, "state", str_getref(side->state));
	if (ret)
		goto err;

	ret = nvl_set_str(nvl, "county", str_getref(side->county));
	if (ret)
		goto err;

	ret = nvl_set_str(nvl, "continent", str_getref(side->continent));
	if (ret)
		goto err;

	ret = nvl_set(nvl, "dxcc", val_getref(side->dxcc));
	if (ret)
		goto err;

	ret = nvl_set(nvl, "itu", val_getref(side->itu));
	if (ret)
		goto err;

	ret = nvl_set(nvl, "cqz", val_getref(side->cqz));
	if (ret)
		goto err;

	ret = nvl_set(nvl, "seq", val_getref(side->seq));
	if (ret)
		goto err;


	return nvl;

err:
	nvl_putref(nvl);

	return ERR_PTR(ret);
}

struct nvlist *qso_pack(const struct qso *qso)
{
	struct nvlist *entry;
	struct nvlist *part;
	int ret;

	entry = nvl_alloc();
	if (IS_ERR(entry))
		return entry;

	/*
	 * store metadata
	 */
	ret = nvl_set_str(entry, "hlog-version", STATIC_STR(version_string));
	if (ret)
		goto err;
	ret = nvl_set_int(entry, "qso-version", QSO_VERSION);
	if (ret)
		goto err;

	/*
	 * store the QSO level info
	 */
	ret = nvl_set_cstr_dup(entry, "uuid", qso->uuid_str);
	if (ret)
		goto err;
	ret = nvl_set_time(entry, "start", &qso->start);
	if (ret)
		goto err;
	ret = nvl_set_time(entry, "end", &qso->end);
	if (ret)
		goto err;
	ret = nvl_set_str(entry, "comment", str_getref(qso->comment));
	if (ret)
		goto err;

	/*
	 * store the tx side
	 */
	part = get_part(&qso->tx);
	if (!part) {
		ret = -ENOMEM;
		goto err;
	}
	ret = nvl_set_nvl(entry, "tx", part);
	if (ret)
		goto err;

	/*
	 * store the rx side
	 */
	part = get_part(&qso->rx);
	if (!part) {
		ret = -ENOMEM;
		goto err;
	}
	ret = nvl_set_nvl(entry, "rx", part);
	if (ret)
		goto err;

	/*
	 * store additional fields
	 */
	ret = nvl_set_nvl(entry, "addl", nvl_getref(qso->additional));
	if (ret)
		goto err;

	return entry;

err:
	nvl_putref(entry);

	return ERR_PTR(ret);
}

static int __unpack_version(struct nvlist *nvl)
{
	uint64_t ver;
	int ret;

	ret = nvl_lookup_int(nvl, "qso-version", &ver);
	if (ret)
		return ret;

	if (ver != QSO_VERSION)
		return -ERANGE;

	ASSERT0(nvl_unset(nvl, "qso-version"));

	return 0;
}

static struct val *__unpack_val(struct nvlist *nvl, const char *name,
				enum val_type vt)
{
	struct val *val;

	val = nvl_lookup_val(nvl, name);
	if (IS_ERR(val))
		return ERR_CAST(val);

	if (val->type == vt) {
		/* nothing to do */
	} else if (val->type == VT_NULL) {
		/* convert to NULL */
		val_putref(val);
		val = NULL;
	} else {
		/* error */
		val_putref(val);
		return ERR_PTR(-ERANGE);
	}

	ASSERT0(nvl_unset(nvl, name));

	return val;
}

static struct str *__unpack_str(struct nvlist *nvl, const char *name)
{
	return val_cast_to_str(__unpack_val(nvl, name, VT_STR));
}

static struct val *__unpack_int(struct nvlist *nvl, const char *name)
{
	return __unpack_val(nvl, name, VT_INT);
}

static int __unpack_datetime(struct time *time, struct nvlist *nvl,
			     const char *name)
{
	struct val **arr;
	size_t nelem;
	int ret;

	ret = nvl_lookup_array(nvl, name, &arr, &nelem);
	if (ret)
		return ret;

	if (nelem != 6)
		return -EINVAL;

#define ONE(idx, member, part) \
	do { \
		if (arr[idx]->type == VT_INT) { \
			time->member = arr[idx]->i; \
		} else if (arr[idx]->type == VT_NULL) { \
			time->member = 0; \
			time->have &= ~(part); \
		} else { \
			return -EINVAL; \
		} \
	} while (0)

	time->have = TIME_PART_DATE | TIME_PART_TIME_HMS;

	ONE(0, year, TIME_PART_DATE);
	ONE(1, month, TIME_PART_DATE);
	ONE(2, day, TIME_PART_DATE);
	ONE(3, hour, TIME_PART_TIME_HM);
	ONE(4, min, TIME_PART_TIME_HM);
	ONE(5, sec, TIME_PART_TIME_S);

#undef ONE

	ASSERT0(nvl_unset(nvl, name));

	return 0;
}

static int __unpack_side(struct qso_side *side, struct nvlist *nvl,
			 const char *name)
{
	struct nvlist *child;
	int ret;

	child = nvl_lookup_nvl(nvl, name);
	if (IS_ERR(child))
		return PTR_ERR(child);

#define S(var, name) \
	do { \
		struct str *str; \
		str = __unpack_str(child, name); \
		if (IS_ERR(str)) { \
			ret = PTR_ERR(str); \
			goto err; \
		} \
		(var) = str; \
	} while (0)
#define I(var, name) \
	do { \
		struct val *val; \
		val = __unpack_int(child, name); \
		if (IS_ERR(val)) { \
			ret = PTR_ERR(val); \
			goto err; \
		} \
		(var) = val; \
	} while (0)

	S(side->station_call, "station");
	S(side->operator_call, "operator");
	I(side->power, "power");
	I(side->freq, "freq");
	S(side->mode, "mode");
	S(side->rst, "rst");
	S(side->name, "name");
	S(side->qth, "qth");
	S(side->grid, "grid");
	S(side->country, "country");
	S(side->state, "state");
	S(side->county, "county");
	S(side->continent, "continent");
	I(side->dxcc, "dxcc");
	I(side->itu, "itu");
	I(side->cqz, "cqz");
	I(side->seq, "seq");

	if (!nvl_numitems(child))
		ASSERT0(nvl_unset(nvl, name));

	return 0;

err:
	nvl_putref(child);

	return ret;
}

static int __unpack_additional(struct qso *qso, struct nvlist *nvl)
{
	struct nvlist *child;

	child = nvl_lookup_nvl(nvl, "addl");
	if (IS_ERR(child)) {
		int ret = PTR_ERR(child);

		return (ret == -ENOENT) ? 0 : ret;
	}

	nvl_putref(qso->additional);
	qso->additional = child;

	ASSERT0(nvl_unset(nvl, "addl"));

	return 0;
}

static int __unpack_other(struct qso *qso, struct nvlist *nvl,
			  const char *prefix)
{
	const struct nvpair *pair;

	nvl_for_each(pair, nvl) {
		const char *name;
		size_t fullname_size;
		char *fullname;
		struct val *val;
		int ret;

		name = nvpair_name(pair);
		val = nvpair_value(pair);

		/* <prefix>-<name><nul> */
		fullname_size = strlen(prefix) + 1 + strlen(name) + 1;
		fullname = alloca(fullname_size);

		snprintf(fullname, fullname_size, "%s-%s", prefix, name);

		switch (val->type) {
			case VT_INT:
			case VT_STR:
			case VT_BOOL:
			case VT_CHAR:
			case VT_NULL:
				ret = nvl_set(qso->additional, fullname, val);
				if (ret)
					return ret;
				break;
			case VT_NVL:
				ret = __unpack_other(qso, val_cast_to_nvl(val),
						     fullname);
				if (ret)
					return ret;
				break;
			case VT_SYM:
			case VT_CONS:
			case VT_BLOB:
			case VT_ARRAY:
				panic("impossible");
		}
	}

	return 0;
}

int qso_unpack(struct qso *qso, struct nvlist *nvl)
{
	struct str *str;
	int ret;

	qso_reset(qso, true);

	/* check that we're dealing with the correct version */
	ret = __unpack_version(nvl);
	if (ret)
		return ret;

	/* ignore version string - it's only useful for debugging */
	nvl_unset(nvl, "hlog-version");

	/* handle per-QSO value: uuid */
	str = __unpack_str(nvl, "uuid");
	if (IS_ERR(str))
		return PTR_ERR(str);

	if (!xuuid_parse(&qso->uuid, str_cstr(str))) {
		str_putref(str);
		return -EINVAL;
	}

	xuuid_unparse(&qso->uuid, qso->uuid_str);

	str_putref(str);

	/* handle per-QSO value: comment */
	str = __unpack_str(nvl, "comment");
	if (IS_ERR(str))
		return PTR_ERR(str);

	qso->comment = str;

	/* handle per-QSO value: start & end dates */
	ret = __unpack_datetime(&qso->start, nvl, "start");
	if (ret)
		return ret;

	ret = __unpack_datetime(&qso->end, nvl, "end");
	if (ret)
		return ret;

	/* handle per-side values: tx */
	ret = __unpack_side(&qso->tx, nvl, "tx");
	if (ret)
		return ret;

	/* handle per-side values: rx */
	ret = __unpack_side(&qso->rx, nvl, "rx");
	if (ret)
		return ret;

	/* handle additional sub-nvlist */
	ret = __unpack_additional(qso, nvl);
	if (ret)
		return ret;

	/* move anything unused to additional */
	return __unpack_other(qso, nvl, "unknown");
}

==> hlog-0.15-rc1.tar.gz: hlog-0.15-rc1/qso_serde.c <==
/*
 * Copyright (c) 2020 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <jeffpc/error.h>
#include <jeffpc/nvl.h>

#include "qso_impl.h"
#include "version.h"

#define QSO_VERSION	1

static int nvl_set_time(struct nvlist *nvl, const char *name,
			const struct time *time)
{
	struct val *arr[6];

	if (time->have & TIME_PART_DATE) {
		arr[0] = VAL_ALLOC_INT(time->year);
		arr[1] = VAL_ALLOC_INT(time->month);
		arr[2] = VAL_ALLOC_INT(time->day);
	} else {
		arr[0] = VAL_ALLOC_NULL();
		arr[1] = VAL_ALLOC_NULL();
		arr[2] = VAL_ALLOC_NULL();
	}

	if (time->have & TIME_PART_TIME_HM) {
		arr[3] = VAL_ALLOC_INT(time->hour);
		arr[4] = VAL_ALLOC_INT(time->min);
	} else {
		arr[3] = VAL_ALLOC_NULL();
		arr[4] = VAL_ALLOC_NULL();
	}

	if (time->have & TIME_PART_TIME_S)
		arr[5] = VAL_ALLOC_INT(time->sec);
	else
		arr[5] = VAL_ALLOC_NULL();

	return nvl_set_array_copy(nvl, name, arr, ARRAY_LEN(arr));
}

static struct nvlist *pack_side(const struct qso_side *side)
{
	struct nvlist *nvl;
	int ret;

	nvl = nvl_alloc();
	if (IS_ERR(nvl))
		return nvl;

	ret = nvl_set_str(nvl, "station", str_getref(side->station_call));
	if (ret)
		goto err;

	ret = nvl_set_str(nvl, "operator", str_getref(side->operator_call));
	if (ret)
		goto err;

	ret = nvl_set(nvl, "power", val_getref(side->power));
	if (ret)
		goto err;

	ret = nvl_set(nvl, "freq", val_getref(side->freq));
	if (ret)
		goto err;

	ret = nvl_set_str(nvl, "mode", str_getref(side->mode));
	if (ret)
		goto err;

	ret = nvl_set_str(nvl, "rst", str_getref(side->rst));
	if (ret)
		goto err;

	ret = nvl_set_str(nvl, "name", str_getref(side->name));
	if (ret)
		goto err;

	ret = nvl_set_str(nvl, "qth", str_getref(side->qth));
	if (ret)
		goto err;

	ret = nvl_set_str(nvl, "grid", str_getref(side->grid));
	if (ret)
		goto err;

	ret = nvl_set_str(nvl, "country", str_getref(side->country));
	if (ret)
		goto err;

	ret = nvl_set_str(nvl, "state", str_getref(side->state));
	if (ret)
		goto err;

	ret = nvl_set_str(nvl, "county", str_getref(side->county));
	if (ret)
		goto err;

	ret = nvl_set_str(nvl, "continent", str_getref(side->continent));
	if (ret)
		goto err;

	ret = nvl_set(nvl, "dxcc", val_getref(side->dxcc));
	if (ret)
		goto err;

	ret = nvl_set(nvl, "itu", val_getref(side->itu));
	if (ret)
		goto err;

	ret = nvl_set(nvl, "cqz", val_getref(side->cqz));
	if (ret)
		goto err;

	ret = nvl_set(nvl, "seq", val_getref(side->seq));
	if (ret)
		goto err;

	return nvl;

err:
	nvl_putref(nvl);

	return ERR_PTR(ret);
}

struct nvlist *qso_pack(const struct qso *qso)
{
	struct nvlist *entry;
	struct nvlist *part;
	int ret;

	entry = nvl_alloc();
	if (IS_ERR(entry))
		return entry;

	/*
	 * store metadata
	 */
	ret = nvl_set_str(entry, "hlog-version", STATIC_STR(version_string));
	if (ret)
		goto err;
	ret = nvl_set_int(entry, "qso-version", QSO_VERSION);
	if (ret)
		goto err;

	/*
	 * store the QSO level info
	 */
	ret = nvl_set_cstr_dup(entry, "uuid", qso->uuid_str);
	if (ret)
		goto err;
	ret = nvl_set_time(entry, "start", &qso->start);
	if (ret)
		goto err;
	ret = nvl_set_time(entry, "end", &qso->end);
	if (ret)
		goto err;
	ret = nvl_set_str(entry, "comment", str_getref(qso->comment));
	if (ret)
		goto err;

	/*
	 * store the tx side
	 */
	part = pack_side(&qso->tx);
	if (!part) {
		ret = -ENOMEM;
		goto err;
	}
	ret = nvl_set_nvl(entry, "tx", part);
	if (ret)
		goto err;

	/*
	 * store the rx side
	 */
	part = pack_side(&qso->rx);
	if (!part) {
		ret = -ENOMEM;
		goto err;
	}
	ret = nvl_set_nvl(entry, "rx", part);
	if (ret)
		goto err;

	/*
	 * store additional fields
	 */
	ret = nvl_set_nvl(entry, "addl", nvl_getref(qso->additional));
	if (ret)
		goto err;

	return entry;

err:
	nvl_putref(entry);

	return ERR_PTR(ret);
}

static int __unpack_version(struct nvlist *nvl)
{
	uint64_t ver;
	int ret;

	ret = nvl_lookup_int(nvl, "qso-version", &ver);
	if (ret)
		return ret;

	if (ver != QSO_VERSION)
		return -ERANGE;

	ASSERT0(nvl_unset(nvl, "qso-version"));

	return 0;
}

static struct val *__unpack_val(struct nvlist *nvl, const char *name,
				enum val_type vt)
{
	struct val *val;

	val = nvl_lookup_val(nvl, name);
	if (IS_ERR(val))
		return ERR_CAST(val);

	if (val->type == vt) {
		/* nothing to do */
	} else if (val->type == VT_NULL) {
		/* convert to NULL */
		val_putref(val);
		val = NULL;
	} else {
		/* error */
		val_putref(val);
		return ERR_PTR(-ERANGE);
	}

	ASSERT0(nvl_unset(nvl, name));

	return val;
}

static struct str *__unpack_str(struct nvlist *nvl, const char *name)
{
	return val_cast_to_str(__unpack_val(nvl, name, VT_STR));
}

static struct val *__unpack_int(struct nvlist *nvl, const char *name)
{
	return __unpack_val(nvl, name, VT_INT);
}

static int __unpack_datetime(struct time *time, struct nvlist *nvl,
			     const char *name)
{
	struct val **arr;
	size_t nelem;
	int ret;

	ret = nvl_lookup_array(nvl, name, &arr, &nelem);
	if (ret)
		return ret;

	if (nelem != 6)
		return -EINVAL;

#define ONE(idx, member, part) \
	do { \
		if (arr[idx]->type == VT_INT) { \
			time->member = arr[idx]->i; \
		} else if (arr[idx]->type == VT_NULL) { \
			time->member = 0; \
			time->have &= ~(part); \
		} else { \
			return -EINVAL; \
		} \
	} while (0)

	time->have = TIME_PART_DATE | TIME_PART_TIME_HMS;

	ONE(0, year, TIME_PART_DATE);
	ONE(1, month, TIME_PART_DATE);
	ONE(2, day, TIME_PART_DATE);
	ONE(3, hour, TIME_PART_TIME_HM);
	ONE(4, min, TIME_PART_TIME_HM);
	ONE(5, sec, TIME_PART_TIME_S);

#undef ONE

	ASSERT0(nvl_unset(nvl, name));

	return 0;
}

static int __unpack_side(struct qso_side *side, struct nvlist *nvl,
			 const char *name)
{
	struct nvlist *child;
	int ret;

	child = nvl_lookup_nvl(nvl, name);
	if (IS_ERR(child))
		return PTR_ERR(child);

#define S(var, name) \
	do { \
		struct str *str; \
		str = __unpack_str(child, name); \
		if (IS_ERR(str)) { \
			ret = PTR_ERR(str); \
			goto err; \
		} \
		(var) = str; \
	} while (0)
#define I(var, name) \
	do { \
		struct val *val; \
		val = __unpack_int(child, name); \
		if (IS_ERR(val)) { \
			ret = PTR_ERR(val); \
			goto err; \
		} \
		(var) = val; \
	} while (0)

	S(side->station_call, "station");
	S(side->operator_call, "operator");
	I(side->power, "power");
	I(side->freq, "freq");
	S(side->mode, "mode");
	S(side->rst, "rst");
	S(side->name, "name");
	S(side->qth, "qth");
	S(side->grid, "grid");
	S(side->country, "country");
	S(side->state, "state");
	S(side->county, "county");
	S(side->continent, "continent");
	I(side->dxcc, "dxcc");
	I(side->itu, "itu");
	I(side->cqz, "cqz");
	I(side->seq, "seq");

	if (!nvl_numitems(child))
		ASSERT0(nvl_unset(nvl, name));

	return 0;

err:
	nvl_putref(child);

	return ret;
}

static int __unpack_additional(struct qso *qso, struct nvlist *nvl)
{
	struct nvlist *child;

	child = nvl_lookup_nvl(nvl, "addl");
	if (IS_ERR(child)) {
		int ret = PTR_ERR(child);

		return (ret == -ENOENT) ? 0 : ret;
	}

	nvl_putref(qso->additional);
	qso->additional = child;

	ASSERT0(nvl_unset(nvl, "addl"));

	return 0;
}

static int __unpack_other(struct qso *qso, struct nvlist *nvl,
			  const char *prefix)
{
	const struct nvpair *pair;

	nvl_for_each(pair, nvl) {
		const char *name;
		size_t fullname_size;
		char *fullname;
		struct val *val;
		int ret;

		name = nvpair_name(pair);
		val = nvpair_value(pair);

		/* <prefix>-<name><nul> */
		fullname_size = strlen(prefix) + 1 + strlen(name) + 1;
		fullname = alloca(fullname_size);

		snprintf(fullname, fullname_size, "%s-%s", prefix, name);

		switch (val->type) {
			case VT_INT:
			case VT_STR:
			case VT_BOOL:
			case VT_CHAR:
			case VT_NULL:
				ret = nvl_set(qso->additional, fullname, val);
				if (ret)
					return ret;
				break;
			case VT_NVL:
				ret = __unpack_other(qso, val_cast_to_nvl(val),
						     fullname);
				if (ret)
					return ret;
				break;
			case VT_SYM:
			case VT_CONS:
			case VT_BLOB:
			case VT_ARRAY:
				panic("impossible");
		}
	}

	return 0;
}

int qso_unpack(struct qso *qso, struct nvlist *nvl)
{
	struct str *str;
	int ret;

	qso_reset(qso, true);

	/* check that we're dealing with the correct version */
	ret = __unpack_version(nvl);
	if (ret)
		return ret;

	/* ignore version string - it's only useful for debugging */
	nvl_unset(nvl, "hlog-version");

	/* handle per-QSO value: uuid */
	str = __unpack_str(nvl, "uuid");
	if (IS_ERR(str))
		return PTR_ERR(str);

	if (!xuuid_parse(&qso->uuid, str_cstr(str))) {
		str_putref(str);
		return -EINVAL;
	}

	xuuid_unparse(&qso->uuid, qso->uuid_str);

	str_putref(str);

	/* handle per-QSO value: comment */
	str = __unpack_str(nvl, "comment");
	if (IS_ERR(str))
		return PTR_ERR(str);

	qso->comment = str;

	/* handle per-QSO value: start & end dates */
	ret = __unpack_datetime(&qso->start, nvl, "start");
	if (ret)
		return ret;

	ret = __unpack_datetime(&qso->end, nvl, "end");
	if (ret)
		return ret;

	/* handle per-side values: tx */
	ret = __unpack_side(&qso->tx, nvl, "tx");
	if (ret)
		return ret;

	/* handle per-side values: rx */
	ret = __unpack_side(&qso->rx, nvl, "rx");
	if (ret)
		return ret;

	/* handle additional sub-nvlist */
	ret = __unpack_additional(qso, nvl);
	if (ret)
		return ret;

	/* move anything unused to additional */
	return __unpack_other(qso, nvl, "unknown");
}

/*
 * Clean up the passed in nvlist before calling qso_unpack, this is useful
 * when we constructed the nvlist from some other source (e.g., an ADIF
 * file).
 */
int qso_unpack_permissive(struct qso *qso, struct nvlist *nvl)
{
	const struct nvpair *pair;
	int ret;

	/* Check that the caller didn't try to guess the version */
	pair = nvl_lookup(nvl, "qso-version");
	if (pair != ERR_PTR(-ENOENT))
		return -EEXIST;

	ret = nvl_set_int(nvl, "qso-version", QSO_VERSION);
	if (ret)
		return ret;

	return qso_unpack(qso, nvl);
}

struct nvlist *qso_init_entry_nvl(void)
{
	static const struct qso empty;
	struct nvlist *nvl, *addl;

	addl = nvl_alloc();
	if (IS_ERR(nvl))
		return addl;

	nvl = qso_pack(&empty);
	if (IS_ERR(nvl)) {
		nvl_putref(addl);
		return nvl;
	}

	/* remove these to avoid confusion later on */
	ASSERT0(nvl_unset(nvl, "hlog-version"));
	ASSERT0(nvl_unset(nvl, "qso-version"));

	/* make addl an nvlist - cannot fail (it exists == no allocs) */
	ASSERT0(nvl_set_nvl(nvl, "addl", addl));

	return nvl;
}
